/*
 * cpg-network-serializer.c
 * This file is part of cpg-network
 *
 * Copyright (C) 2011 - Jesse van den Kieboom
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include "cpg-network-serializer.h"
#include "cpg-enum-types.h"
#include "cpg-function-polynomial.h"
#include "cpg-network-xml.h"
#include "cpg-import.h"
#include "cpg-input-file.h"
#include "cpg-annotatable.h"
#include "cpg-layoutable.h"

#include <libxml/tree.h>

/**
 * SECTION:cpg-network-serializer
 * @short_description: Network to XML serializer
 *
 * This can be used to serialize a #CpgNetwork to XML.
 *
 */
#define CPG_NETWORK_SERIALIZER_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE((object), CPG_TYPE_NETWORK_SERIALIZER, CpgNetworkSerializerPrivate))

struct _CpgNetworkSerializerPrivate
{
	CpgNetwork *network;
	CpgGroup *root;
	xmlDocPtr doc;
};

G_DEFINE_TYPE (CpgNetworkSerializer, cpg_network_serializer, G_TYPE_OBJECT)

enum
{
	PROP_0,
	PROP_NETWORK,
	PROP_ROOT
};

static void
cpg_network_serializer_finalize (GObject *object)
{
	G_OBJECT_CLASS (cpg_network_serializer_parent_class)->finalize (object);
}

static void
cpg_network_serializer_dispose (GObject *object)
{
	CpgNetworkSerializer *serializer = CPG_NETWORK_SERIALIZER (object);

	if (serializer->priv->network)
	{
		g_object_unref (serializer->priv->network);
		serializer->priv->network = NULL;
	}

	if (serializer->priv->root)
	{
		g_object_unref (serializer->priv->root);
		serializer->priv->root = NULL;
	}

	G_OBJECT_CLASS (cpg_network_serializer_parent_class)->dispose (object);
}

static void
cpg_network_serializer_set_property (GObject      *object,
                                     guint         prop_id,
                                     const GValue *value,
                                     GParamSpec   *pspec)
{
	CpgNetworkSerializer *self = CPG_NETWORK_SERIALIZER (object);

	switch (prop_id)
	{
		case PROP_NETWORK:
			self->priv->network = g_value_dup_object (value);
		break;
		case PROP_ROOT:
			self->priv->root = g_value_dup_object (value);
		break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
cpg_network_serializer_get_property (GObject    *object,
                                     guint       prop_id,
                                     GValue     *value,
                                     GParamSpec *pspec)
{
	CpgNetworkSerializer *self = CPG_NETWORK_SERIALIZER (object);

	switch (prop_id)
	{
		case PROP_NETWORK:
			g_value_set_object (value, self->priv->network);
		break;
		case PROP_ROOT:
			g_value_set_object (value, self->priv->root);
		break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
cpg_network_serializer_class_init (CpgNetworkSerializerClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = cpg_network_serializer_finalize;
	object_class->dispose = cpg_network_serializer_dispose;

	object_class->get_property = cpg_network_serializer_get_property;
	object_class->set_property = cpg_network_serializer_set_property;

	g_type_class_add_private (object_class, sizeof(CpgNetworkSerializerPrivate));

	/**
	 * CpgSerializer:network:
	 *
	 * The #CpgNetwork to serialize.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_NETWORK,
	                                 g_param_spec_object ("network",
	                                                      "Network",
	                                                      "Network",
	                                                      CPG_TYPE_NETWORK,
	                                                      G_PARAM_READWRITE |
	                                                      G_PARAM_CONSTRUCT_ONLY));

	/**
	 * CpgSerializer:root:
	 *
	 * The #CpgGroup to serialize
	 *
	 */
	g_object_class_install_property (object_class,
	                                 PROP_ROOT,
	                                 g_param_spec_object ("root",
	                                                      "Root",
	                                                      "Root",
	                                                      CPG_TYPE_GROUP,
	                                                      G_PARAM_READWRITE |
	                                                      G_PARAM_CONSTRUCT_ONLY));
}

static void
cpg_network_serializer_init (CpgNetworkSerializer *self)
{
	self->priv = CPG_NETWORK_SERIALIZER_GET_PRIVATE (self);
}

/**
 * cpg_network_serializer_new:
 * @network: A #CpgNetwork
 * @root: A #CpgGroup
 *
 * Create a new serializer for @network. When serialized, only objects
 * contained in @root will be serialized. This can be useful to serialize
 * only parts of the network. If @root is %NULL, all objects will be
 * serialized.
 *
 * Returns: A #CpgNetworkSerializer
 *
 **/
CpgNetworkSerializer *
cpg_network_serializer_new (CpgNetwork *network,
                            CpgGroup   *root)
{
	return g_object_new (CPG_TYPE_NETWORK_SERIALIZER,
	                     "network", network,
	                     "root", root,
	                     NULL);
}

extern int xmlIndentTreeOutput;

static void group_to_xml (CpgNetworkSerializer *serializer, xmlNodePtr root, CpgGroup *group);

static gboolean
property_matches_template (CpgProperty  *property,
                           GSList const *templates)
{
	while (templates)
	{
		CpgProperty *tprop;

		tprop = cpg_object_get_property (templates->data,
		                                 cpg_property_get_name (property));

		if (tprop)
		{
			return cpg_property_equal (property, tprop);
		}

		templates = g_slist_next (templates);
	}

	return FALSE;
}

static void
export_flags (xmlNodePtr   node,
              CpgProperty *property)
{
	CpgPropertyFlags add_flags;
	CpgPropertyFlags remove_flags;
	GFlagsClass *klass;
	guint i;
	CpgObject *templ;

	gboolean flags_attr;

	flags_attr = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (property),
	                                                 CPG_NETWORK_XML_PROPERTY_FLAGS_ATTRIBUTE));

	add_flags = cpg_property_get_flags (property);
	remove_flags = CPG_PROPERTY_FLAG_NONE;

	templ = cpg_object_get_property_template (cpg_property_get_object (property),
	                                          property,
	                                          FALSE);

	if (templ)
	{
		/* See how to change */
		CpgProperty *orig;

		orig = cpg_object_get_property (templ, cpg_property_get_name (property));

		remove_flags = ~add_flags & cpg_property_get_flags (orig);
		add_flags = add_flags & ~cpg_property_get_flags (orig);
	}

	if (flags_attr)
	{
		gchar *s = cpg_property_flags_to_string (add_flags, remove_flags);
		xmlNewProp (node, (xmlChar *)"flags", (xmlChar *)s);
		g_free (s);

		return;
	}

	klass = g_type_class_ref (CPG_TYPE_PROPERTY_FLAGS);

	for (i = 0; i < klass->n_values; ++i)
	{
		GFlagsValue *value = &(klass->values[i]);

		if (value->value == 0)
		{
			continue;
		}

		if ((add_flags & value->value) == value->value)
		{
			xmlNewProp (node,
			            (xmlChar *)value->value_nick,
			            (xmlChar *)"yes");
		}
		else if ((remove_flags & value->value) == value->value)
		{
			xmlNewProp (node,
			            (xmlChar *)value->value_nick,
			            (xmlChar *)"no");
		}
	}
}

static void
restore_comment (CpgNetworkSerializer *serializer,
                 xmlNodePtr            parent,
                 GObject              *object)
{
	gchar *comment;
	gchar *with_spaces;

	if (CPG_IS_ANNOTATABLE (object))
	{
		comment = cpg_annotatable_get_annotation (CPG_ANNOTATABLE (object));
	}
	else
	{
		comment = g_strdup (g_object_get_data (object,
		                                       CPG_NETWORK_XML_COMMENT_DATA_KEY));
	}

	if (comment == NULL)
	{
		return;
	}

	with_spaces = g_strconcat (" ", comment, " ", NULL);

	xmlNodePtr node = xmlNewDocComment (serializer->priv->doc,
	                                    (xmlChar *)with_spaces);

	g_free (with_spaces);

	xmlAddChild (parent, node);
}

static void
properties_to_xml (CpgNetworkSerializer *serializer,
                   xmlNodePtr            parent,
                   CpgObject            *object,
                   GSList const         *props)
{
	GSList *item;
	GSList *properties;
	GSList const *templates;

	if (props)
	{
		properties = g_slist_copy ((GSList *)props);
	}
	else
	{
		properties = cpg_object_get_properties (object);
	}

	templates = cpg_object_get_applied_templates (object);

	for (item = properties; item; item = g_slist_next (item))
	{
		CpgProperty *property = item->data;
		CpgExpression *cons;

		if (cpg_property_get_object (property) != object)
		{
			continue;
		}

		CpgExpression *expression = cpg_property_get_expression (property);

		// Check if property is different from template
		if (property_matches_template (property, templates))
		{
			continue;
		}

		restore_comment (serializer, parent, G_OBJECT (property));

		xmlNodePtr node = xmlNewDocNode (serializer->priv->doc,
		                                 NULL,
		                                 (xmlChar *)"property",
		                                 NULL);
		xmlNewProp (node,
		            (xmlChar *)"name",
		            (xmlChar *)cpg_property_get_name (property));

		export_flags (node, property);

		gchar const *expr = cpg_expression_get_as_string (expression);

		if (expr && *expr)
		{
			xmlNodePtr text = xmlNewDocText (serializer->priv->doc,
			                                 (xmlChar *)expr);

			xmlAddChild (node, text);
		}

		cons = cpg_property_get_constraint (property);

		if (cons)
		{
			xmlNewProp (node,
			            (xmlChar *)"constraint",
			            (xmlChar *)cpg_expression_get_as_string (cons));
		}

		xmlAddChild (parent, node);
	}

	g_slist_free (properties);
}

static gchar *
template_path (CpgNetwork *network,
               CpgObject *orig,
               CpgObject *template)
{
	CpgObject *parent = cpg_object_get_parent (orig);
	CpgObject *shared_root = template;

	GString *ret = g_string_new ("");
	gboolean first = TRUE;

	CpgObject *tg;

	tg = CPG_OBJECT (cpg_network_get_template_group (network));

	do
	{
		CpgObject *shared_parent = cpg_object_get_parent (shared_root);

		if (shared_root == tg)
		{
			break;
		}

		if (shared_parent != NULL)
		{
			if (!first)
			{
				g_string_prepend (ret, ".");
			}

			g_string_prepend (ret, cpg_object_get_id (shared_root));

			first = FALSE;
		}

		shared_root = shared_parent;
	} while (shared_root && parent != shared_root);

	return g_string_free (ret, FALSE);
}

static gboolean
should_write_layout (CpgObject *object)
{
	CpgLayoutable *layoutable;
	CpgLayoutable *last;
	gint x;
	gint y;
	gint tx;
	gint ty;
	GSList *templates;
	GSList *item;

	if (!CPG_IS_LAYOUTABLE (object))
	{
		return FALSE;
	}

	layoutable = CPG_LAYOUTABLE (object);

	if (!cpg_layoutable_supports_location (layoutable) ||
	    !cpg_layoutable_get_has_location (layoutable))
	{
		return FALSE;
	}

	/* Check if the layout was not applied by some template */
	templates = cpg_group_get_auto_templates_for_child (CPG_GROUP (cpg_object_get_parent (object)),
	                                                    object);

	last = NULL;

	for (item = templates; item; item = g_slist_next (item))
	{
		if (CPG_IS_LAYOUTABLE (item->data) &&
		    cpg_layoutable_supports_location (item->data) &&
		    cpg_layoutable_get_has_location (item->data))
		{
			last = item->data;
		}
	}

	g_slist_free (templates);

	if (!last)
	{
		return TRUE;
	}

	cpg_layoutable_get_location (layoutable, &x, &y);
	cpg_layoutable_get_location (last, &tx, &ty);

	return x != tx || y != ty;
}

static void
add_layout (CpgObject  *object,
            xmlNodePtr  ptr)
{
	if (should_write_layout (object))
	{
		gchar *pos;
		gint x;
		gint y;

		cpg_layoutable_get_location (CPG_LAYOUTABLE (object), &x, &y);

		pos = g_strdup_printf ("%d", x);
		xmlNewProp (ptr, (xmlChar *)"x", (xmlChar *)pos);
		g_free (pos);

		pos = g_strdup_printf ("%d", y);
		xmlNewProp (ptr, (xmlChar *)"y", (xmlChar *)pos);
		g_free (pos);
	}
}

static xmlNodePtr
object_to_xml (CpgNetworkSerializer *serializer,
               xmlNodePtr            parent,
               CpgObject            *object,
               gchar const          *name,
               GSList const         *properties)
{
	restore_comment (serializer, parent, G_OBJECT (object));

	xmlNodePtr ptr = xmlNewDocNode (serializer->priv->doc,
	                                NULL,
	                                (xmlChar *)name,
	                                NULL);

	xmlNewProp (ptr, (xmlChar *)"id", (xmlChar *)cpg_object_get_id (object));
	xmlAddChild (parent, ptr);

	add_layout (object, ptr);

	GSList const *templates = cpg_object_get_applied_templates (object);
	GSList *inherited = cpg_group_get_auto_templates_for_child (CPG_GROUP (cpg_object_get_parent (object)),
	                                                            object);
	GPtrArray *refs = g_ptr_array_new ();

	while (templates)
	{
		/* only apply templates that are not inherited */
		if (!g_slist_find (inherited, templates->data))
		{
			gchar *path = template_path (serializer->priv->network,
			                             object,
			                             templates->data);

			g_ptr_array_add (refs, path);
		}

		templates = g_slist_next (templates);
	}

	g_slist_free (inherited);

	g_ptr_array_add (refs, NULL);

	gchar **refs_ptr = (gchar **)g_ptr_array_free (refs, FALSE);

	if (refs_ptr && *refs_ptr)
	{
		gchar *joined = g_strjoinv (", ", refs_ptr);
		xmlNewProp (ptr, (xmlChar *)"ref", (xmlChar *)joined);
		g_free (joined);
	}

	g_strfreev (refs_ptr);

	properties_to_xml (serializer,
	                   ptr,
	                   object,
	                   properties);

	return ptr;
}

static xmlNodePtr
state_to_xml (CpgNetworkSerializer *serializer,
              xmlNodePtr            parent,
              CpgObject            *state)
{
	return object_to_xml (serializer, parent, state, "state", NULL);
}

static gboolean
action_matches_template (CpgLinkAction *action,
                         GSList const  *templates)
{
	gchar const *p1 = cpg_link_action_get_target (action);
	CpgExpression *e1 = cpg_link_action_get_equation (action);

	while (templates)
	{
		CpgObject *template = templates->data;
		templates = g_slist_next (templates);

		if (!CPG_IS_LINK (template))
		{
			continue;
		}

		GSList const *actions = cpg_link_get_actions (CPG_LINK (template));

		while (actions)
		{
			CpgLinkAction *other = actions->data;

			gchar const *p2 = cpg_link_action_get_target (other);
			CpgExpression *e2 = cpg_link_action_get_equation (other);

			if (g_strcmp0 (p1, p2) == 0)
			{
				return cpg_expression_equal (e1, e2);
			}

			actions = g_slist_next (actions);
		}
	}

	return FALSE;
}

static xmlNodePtr
link_to_xml (CpgNetworkSerializer *serializer,
             xmlNodePtr            parent,
             CpgLink              *link)
{
	CpgObject *from = cpg_link_get_from (link);
	CpgObject *to = cpg_link_get_to (link);

	xmlNodePtr node = object_to_xml (serializer,
	                                 parent,
	                                 CPG_OBJECT (link),
	                                 "link",
	                                 NULL);

	if (from != NULL)
	{
		xmlNewProp (node,
		            (xmlChar *)"from",
		            (xmlChar *)cpg_object_get_id (from));
	}

	if (to != NULL)
	{
		xmlNewProp (node,
		            (xmlChar *)"to",
		            (xmlChar *)cpg_object_get_id (to));
	}

	// Link actions
	GSList const *item;
	GSList const *templates = cpg_object_get_applied_templates (CPG_OBJECT (link));

	for (item = cpg_link_get_actions (link); item; item = g_slist_next (item))
	{
		CpgLinkAction *action = item->data;

		if (action_matches_template (action, templates))
		{
			continue;
		}

		restore_comment (serializer, node, G_OBJECT (action));

		xmlNodePtr ac = xmlNewDocNode (serializer->priv->doc,
		                               NULL,
		                               (xmlChar *)"action",
		                               NULL);
		xmlNewProp (ac,
		            (xmlChar *)"target",
		            (xmlChar *)cpg_link_action_get_target (action));

		gchar const *expr = cpg_expression_get_as_string (cpg_link_action_get_equation (action));

		if (expr && *expr)
		{
			xmlNodePtr text = xmlNewDocText (serializer->priv->doc,
		                                 (xmlChar *)expr);

			xmlAddChild (ac, text);
		}

		xmlAddChild (node, ac);
	}

	return node;
}

static void
function_to_xml (CpgNetworkSerializer *serializer,
                 xmlNodePtr            parent,
                 CpgFunction          *func)
{
	xmlNodePtr funcn;

	funcn = xmlNewDocNode (serializer->priv->doc, NULL, (xmlChar *)"function", NULL);

	xmlNewProp (funcn,
	            (xmlChar *)"name",
	            (xmlChar *)cpg_object_get_id (CPG_OBJECT (func)));

	xmlAddChild (parent, funcn);

	/* Create expression element */
	CpgExpression *expression = cpg_function_get_expression (func);

	if (expression)
	{
		restore_comment (serializer, funcn, G_OBJECT (expression));

		xmlNodePtr exprn = xmlNewDocNode (serializer->priv->doc,
		                                  NULL,
		                                  (xmlChar *)"expression",
		                                  NULL);

		gchar const *expr = cpg_expression_get_as_string (expression);

		if (expr && *expr)
		{
			xmlNodePtr text = xmlNewDocText (serializer->priv->doc,
			                                 (xmlChar *)expr);

			xmlAddChild (exprn, text);
		}

		xmlAddChild (funcn, exprn);
	}

	/* Create argument elements */
	GList const *args = cpg_function_get_arguments (func);
	GList const *argitem;

	for (argitem = args; argitem; argitem = g_list_next (argitem))
	{
		CpgFunctionArgument *argument = argitem->data;

		restore_comment (serializer, funcn, G_OBJECT (argument));

		xmlNodePtr argn = xmlNewDocNode (serializer->priv->doc,
		                                 NULL,
		                                 (xmlChar *)"argument",
		                                 NULL);
		xmlNodePtr text = xmlNewDocText (serializer->priv->doc,
		                                 (xmlChar *)cpg_function_argument_get_name (argument));

		xmlAddChild (argn, text);

		if (cpg_function_argument_get_optional (argument))
		{
			CpgExpression *expr;

			expr = cpg_function_argument_get_default_value (argument);
			xmlNewProp (argn, (xmlChar *)"default", (xmlChar *)cpg_expression_get_as_string (expr));
		}

		if (!cpg_function_argument_get_explicit (argument))
		{
			xmlNewProp (argn, (xmlChar *)"implicit", (xmlChar *)"yes");
		}

		xmlAddChild (funcn, argn);
	}

	add_layout (CPG_OBJECT (func), funcn);
}

static void
function_polynomial_to_xml (CpgNetworkSerializer  *serializer,
                            xmlNodePtr             parent,
                            CpgFunctionPolynomial *func)
{
	xmlNodePtr funcn;

	funcn = xmlNewDocNode (serializer->priv->doc, NULL, (xmlChar *)"polynomial", NULL);

	xmlNewProp (funcn,
	            (xmlChar *)"name",
	            (xmlChar *)cpg_object_get_id (CPG_OBJECT (func)));

	xmlAddChild (parent, funcn);

	/* Create pieces */
	GSList const *pieces = cpg_function_polynomial_get_pieces (func);

	while (pieces)
	{
		CpgFunctionPolynomialPiece *piece = (CpgFunctionPolynomialPiece *)pieces->data;

		gchar beginPtr[G_ASCII_DTOSTR_BUF_SIZE];
		gchar endPtr[G_ASCII_DTOSTR_BUF_SIZE];

		g_ascii_dtostr (beginPtr,
		                G_ASCII_DTOSTR_BUF_SIZE,
		                cpg_function_polynomial_piece_get_begin (piece));

		g_ascii_dtostr (endPtr,
		                G_ASCII_DTOSTR_BUF_SIZE,
		                cpg_function_polynomial_piece_get_end (piece));

		xmlNodePtr piecen = xmlNewDocNode (serializer->priv->doc, NULL, (xmlChar *)"piece", NULL);
		xmlNewProp (piecen, (xmlChar *)"begin", (xmlChar *)beginPtr);
		xmlNewProp (piecen, (xmlChar *)"end", (xmlChar *)endPtr);

		GString *str = g_string_new ("");
		guint num;
		guint i;
		gdouble const *coefficients;

		coefficients = cpg_function_polynomial_piece_get_coefficients (piece,
		                                                               &num);

		for (i = 0; i < num; ++i)
		{
			gchar coefPtr[G_ASCII_DTOSTR_BUF_SIZE];

			if (i != 0)
			{
				g_string_append (str, ", ");
			}

			g_ascii_dtostr (coefPtr, G_ASCII_DTOSTR_BUF_SIZE, coefficients[i]);
			g_string_append (str, coefPtr);
		}

		xmlNodePtr text = xmlNewDocText (serializer->priv->doc,
		                                 (xmlChar *)str->str);

		g_string_free (str, TRUE);

		xmlAddChild (piecen, text);
		xmlAddChild (funcn, piecen);

		pieces = g_slist_next (pieces);
	}

	add_layout (CPG_OBJECT (func), funcn);
}

static gboolean
skip_object (CpgObject *object)
{
	GSList *templates;

	if (!cpg_object_get_parent (object))
	{
		return FALSE;
	}

	if (cpg_object_get_auto_imported (object))
	{
		return TRUE;
	}

	if (should_write_layout (object))
	{
		return FALSE;
	}

	/* Check if it was not solely instantiated from templates */
	templates = cpg_group_get_auto_templates_for_child (CPG_GROUP (cpg_object_get_parent (object)),
	                                                    object);

	if (!templates)
	{
		return FALSE;
	}

	g_slist_free (templates);

	CpgObject *dummy = g_object_new (G_TYPE_FROM_INSTANCE (object),
	                                 "id", cpg_object_get_id (object),
	                                 NULL);

	_cpg_object_set_parent (dummy, cpg_object_get_parent (object));

	GSList const *item;

	for (item = cpg_object_get_applied_templates (object); item; item = g_slist_next (item))
	{
		cpg_object_apply_template (dummy, item->data, NULL);
	}

	if (cpg_object_equal (object, dummy))
	{
		return TRUE;
	}

	g_object_unref (dummy);
	return FALSE;
}

static void
import_to_xml (CpgNetworkSerializer *serializer,
               xmlNodePtr            root,
               CpgImport            *import)
{
	if (cpg_modifiable_get_modified (CPG_MODIFIABLE (import)))
	{
		group_to_xml (serializer, root, CPG_GROUP (import));
		return;
	}

	restore_comment (serializer, root, G_OBJECT (import));

	xmlNodePtr node = xmlNewDocNode (serializer->priv->doc, NULL, (xmlChar *)"import", NULL);

	xmlNewProp (node, (xmlChar *)"id", (xmlChar *)cpg_object_get_id (CPG_OBJECT (import)));

	add_layout (CPG_OBJECT (import), node);

	GFile *network_file = cpg_network_get_file (serializer->priv->network);
	GFile *import_file = cpg_import_get_file (import);

	gchar *path = NULL;

	if (network_file)
	{
		GFile *parent;

		parent = g_file_get_parent (network_file);

		path = g_file_get_relative_path (parent, import_file);

		g_object_unref (network_file);
		g_object_unref (parent);
	}

	if (!path)
	{
		gchar *cwd = g_get_current_dir ();
		GFile *cwd_file = g_file_new_for_path (cwd);
		g_free (cwd);

		path = g_file_get_relative_path (cwd_file, import_file);
		g_object_unref (cwd_file);
	}

	if (!path)
	{
		path = g_file_get_path (import_file);
	}

	g_object_unref (import_file);

	xmlNodePtr text = xmlNewDocText (serializer->priv->doc,
	                                 (xmlChar *)path);

	xmlAddChild (node, text);
	g_free (path);

	xmlAddChild (root, node);
}

static void
input_file_to_xml (CpgNetworkSerializer *serializer,
                   xmlNodePtr            parent,
                   CpgInputFile         *input)
{
	xmlNodePtr node;
	GSList *properties;
	gchar **columns;
	gchar **item;

	properties = cpg_object_get_properties (CPG_OBJECT (input));

	columns = cpg_input_file_get_columns (input);

	for (item = columns; *item; ++item)
	{
		properties = g_slist_remove (properties,
		                             cpg_object_get_property (CPG_OBJECT (input), *item));
	}

	g_strfreev (columns);

	node = object_to_xml (serializer,
	                      parent,
	                      CPG_OBJECT (input),
	                      "input-file",
	                      properties);

	g_slist_free (properties);

	/* File */
	GFile *file;
	file = cpg_input_file_get_file (input);

	if (file)
	{
		GFile *network_file;
		GFile *parent;

		network_file = cpg_network_get_file (serializer->priv->network);
		gchar *path = NULL;

		if (network_file)
		{
			parent = g_file_get_parent (network_file);

			path = g_file_get_relative_path (parent, file);

			g_object_unref (parent);
			g_object_unref (network_file);
		}

		if (!path)
		{
			path = g_file_get_path (file);
		}

		if (node->children)
		{
			xmlNewProp (node,
			            (xmlChar *)"filename",
			            (xmlChar *)path);
		}
		else
		{
			xmlNodePtr text;

			text = xmlNewDocText (serializer->priv->doc,
			                      (xmlChar *)path);

			xmlAddChild (node, text);
		}

		g_free (path);
		g_object_unref (file);
	}

	/* Repeat */
	if (cpg_input_file_get_repeat (input))
	{
		xmlNewProp (node, (xmlChar *)"repeat", (xmlChar *)"yes");
	}

	/* Time column */
	gint time_column;
	gboolean isset;

	time_column = cpg_input_file_get_time_column (input, &isset);

	if (isset)
	{
		gchar *ptr = g_strdup_printf ("%d", time_column);
		xmlNewProp (node, (xmlChar *)"time-column", (xmlChar *)ptr);
		g_free (ptr);
	}
}

static void
any_object_to_xml (CpgNetworkSerializer *serializer,
                   xmlNodePtr            root,
                   CpgObject            *object)
{
	/* Check if this object can be omitted because it's covered fully
	   by its fake template */
	if (skip_object (object))
	{
		return;
	}

	if (CPG_IS_IMPORT (object))
	{
		import_to_xml (serializer, root, CPG_IMPORT (object));
	}
	else if (CPG_IS_GROUP (object))
	{
		group_to_xml (serializer, root, CPG_GROUP (object));
	}
	else if (CPG_IS_LINK (object))
	{
		link_to_xml (serializer, root, CPG_LINK (object));
	}
	else if (CPG_IS_INPUT_FILE (object))
	{
		input_file_to_xml (serializer, root, CPG_INPUT_FILE (object));
	}
	else if (CPG_IS_FUNCTION_POLYNOMIAL (object))
	{
		function_polynomial_to_xml (serializer, root, CPG_FUNCTION_POLYNOMIAL (object));
	}
	else if (CPG_IS_FUNCTION (object))
	{
		function_to_xml (serializer, root, CPG_FUNCTION (object));
	}
	else
	{
		state_to_xml (serializer, root, object);
	}
}

static gboolean
check_proxy_template (CpgObject *object,
                      CpgObject *proxy)
{
	GSList const *templates = cpg_object_get_applied_templates (object);

	while (templates)
	{
		if (CPG_IS_GROUP (templates->data))
		{
			CpgObject *other_proxy = cpg_group_get_proxy (templates->data);

			if (other_proxy != NULL &&
			    g_strcmp0 (cpg_object_get_id (proxy),
			               cpg_object_get_id (other_proxy)) == 0)
			{
				return TRUE;
			}
		}

		templates = g_slist_next (templates);
	}

	return FALSE;
}

static void
group_children_to_xml (CpgNetworkSerializer *serializer,
                       xmlNodePtr            group_node,
                       CpgGroup             *group)
{
	GSList const *children = cpg_group_get_children (group);
	GSList *links = NULL;

	while (children)
	{
		if (CPG_IS_LINK (children->data))
		{
			links = g_slist_prepend (links, children->data);
		}
		else
		{
			any_object_to_xml (serializer, group_node, children->data);
		}

		children = g_slist_next (children);
	}

	GSList *item;
	links = g_slist_reverse (links);

	for (item = links; item; item = g_slist_next (item))
	{
		any_object_to_xml (serializer, group_node, item->data);
	}

	g_slist_free (links);
}

static gboolean
group_interface_is_template (CpgGroup    *group,
                             gchar const *name)
{
	CpgPropertyInterface *iface;
	GSList const *templates;
	gboolean ret = FALSE;
	gchar const *my_child_name;
	gchar const *my_property_name;

	iface = cpg_group_get_property_interface (group);

	my_child_name = cpg_property_interface_lookup_child_name (iface, name);
	my_property_name = cpg_property_interface_lookup_property_name (iface, name);

	templates = cpg_object_get_applied_templates (CPG_OBJECT (group));

	while (templates)
	{
		CpgGroup *template;
		CpgPropertyInterface *template_iface;
		gchar const *child_name;
		gchar const *property_name;

		template = templates->data;

		template_iface = cpg_group_get_property_interface (template);

		child_name = cpg_property_interface_lookup_child_name (template_iface, name);
		property_name = cpg_property_interface_lookup_property_name (template_iface, name);

		if (g_strcmp0 (my_child_name, child_name) == 0 &&
		    g_strcmp0 (my_property_name, property_name) == 0)
		{
			ret = TRUE;
			break;
		}

		templates = g_slist_next (templates);
	}

	return ret;
}

static gboolean
group_interface_is_proxy (CpgGroup    *group,
                          gchar const *name)
{
	CpgObject *proxy;
	gchar const *child_name;
	CpgPropertyInterface *iface;

	/* Check if the interface is an automatically generated interface from
	   a property on the proxy object */
	proxy = cpg_group_get_proxy (group);

	if (!proxy)
	{
		return FALSE;
	}

	iface = cpg_group_get_property_interface (group);
	child_name = cpg_property_interface_lookup_child_name (iface, name);

	return g_strcmp0 (cpg_object_get_id (proxy), child_name) == 0;
}

static gchar **
find_non_template_interfaces (CpgGroup *group)
{
	GPtrArray *ret;
	gchar **names;
	gchar **ptr;
	CpgPropertyInterface *iface;

	ret = g_ptr_array_new ();

	iface = cpg_group_get_property_interface (group);
	names = cpg_property_interface_get_names (iface);

	for (ptr = names; ptr && *ptr; ++ptr)
	{
		if (group_interface_is_template (group, *ptr))
		{
			continue;
		}

		if (group_interface_is_proxy (group, *ptr))
		{
			continue;
		}

		g_ptr_array_add (ret, g_strdup (*ptr));
	}

	g_ptr_array_add (ret, NULL);
	g_strfreev (names);

	return (gchar **)g_ptr_array_free (ret, FALSE);
}

static void
group_interface_to_xml (CpgNetworkSerializer *serializer,
                        xmlNodePtr            group_node,
                        CpgGroup             *group)
{
	CpgPropertyInterface *iface;
	gchar **names;
	gchar **ptr;
	xmlNodePtr parent;

	iface = cpg_group_get_property_interface (group);

	names = find_non_template_interfaces (group);

	if (names && *names)
	{
		parent = xmlNewDocNode (serializer->priv->doc, NULL, (xmlChar *)"interface", NULL);
		xmlAddChild (group_node, parent);
	}

	for (ptr = names; ptr && *ptr; ++ptr)
	{
		gchar const *child_name;
		gchar const *property_name;
		xmlNodePtr node;
		xmlNodePtr text;

		child_name = cpg_property_interface_lookup_child_name (iface, *ptr);
		property_name = cpg_property_interface_lookup_property_name (iface, *ptr);

		node = xmlNewDocNode (serializer->priv->doc, NULL, (xmlChar *)"property", NULL);
		xmlNewProp (node, (xmlChar *)"name", (xmlChar *)*ptr);
		xmlNewProp (node, (xmlChar *)"child", (xmlChar *)child_name);

		text = xmlNewDocText (serializer->priv->doc,
		                      (xmlChar *)property_name);

		xmlAddChild (node, text);
		xmlAddChild (parent, node);
	}

	g_strfreev (names);
}

static void
group_to_xml (CpgNetworkSerializer *serializer,
              xmlNodePtr            root,
              CpgGroup             *group)
{
	xmlNodePtr group_node;

	if (!CPG_IS_NETWORK (group))
	{
		group_node = object_to_xml (serializer,
		                            root,
		                            CPG_OBJECT (group),
		                            "group",
		                            NULL);

		CpgObject *proxy = cpg_group_get_proxy (group);

		if (proxy != NULL &&
		    !check_proxy_template (CPG_OBJECT (group), proxy))
		{
			xmlNewProp (group_node,
			            (xmlChar *)"proxy",
			            (xmlChar *)cpg_object_get_id (proxy));
		}
	}
	else
	{
		group_node = root;
	}

	group_interface_to_xml (serializer, group_node, group);

	group_children_to_xml (serializer, group_node, group);
}

static void
write_config (CpgNetworkSerializer *serializer,
              xmlNodePtr            nnetwork)
{
	CpgIntegrator *integrator = cpg_network_get_integrator (serializer->priv->network);

	if (integrator != NULL)
	{
		xmlNewProp (nnetwork,
		            (xmlChar *)"integrator",
		            (xmlChar *)cpg_object_get_id (CPG_OBJECT (integrator)));
	}
}

static void
write_extra_nodes (CpgNetworkSerializer *serializer)
{
	/* Restore additional xml nodes from network... */
	xmlDocPtr doc;
	xmlNodePtr root;
	xmlNodePtr write_root;

	doc = g_object_get_data (G_OBJECT (serializer->priv->network),
	                         CPG_NETWORK_XML_EXTRA_DATA_KEY);

	if (!doc)
	{
		return;
	}

	root = xmlDocGetRootElement (doc);
	xmlNodePtr child = root->children;

	write_root = xmlDocGetRootElement (serializer->priv->doc);

	while (child)
	{
		xmlNodePtr cp;

		cp = xmlDocCopyNode (child, serializer->priv->doc, 1);

		if (cp)
		{
			xmlAddChild (write_root, cp);
		}

		child = child->next;
	}
}

/**
 * cpg_network_serializer_serialize:
 * @serializer: A #CpgNetworkSerializer
 * @stream: A #GOutputStream
 * @error: A #GError
 *
 * Serialize a network to an output stream.
 *
 * Returns: %TRUE if the serialization was successful, %FALSE otherwise
 *
 **/
gboolean
cpg_network_serializer_serialize (CpgNetworkSerializer  *serializer,
                                  GOutputStream         *stream,
                                  GError               **error)
{
	g_return_val_if_fail (CPG_IS_NETWORK_SERIALIZER (serializer), FALSE);
	g_return_val_if_fail (G_IS_OUTPUT_STREAM (stream), FALSE);

	xmlDocPtr doc = xmlNewDoc ((xmlChar *)"1.0");
	xmlNodePtr root = xmlNewDocNode (doc, NULL, (xmlChar *)"cpg", NULL);

	xmlDocSetRootElement (doc, root);

	xmlNodePtr nnetwork = xmlNewDocNode (doc, NULL, (xmlChar *)"network", NULL);

	restore_comment (serializer, root, G_OBJECT (serializer->priv->network));

	xmlAddChild (root, nnetwork);


	serializer->priv->doc = doc;

	write_config (serializer, nnetwork);

	// Globals
	GSList *properties = cpg_object_get_properties (CPG_OBJECT (serializer->priv->network));

	if (properties)
	{
		xmlNodePtr gbl = xmlNewDocNode (doc, NULL, (xmlChar *)"globals", NULL);
		xmlAddChild (nnetwork, gbl);

		properties_to_xml (serializer, gbl, CPG_OBJECT (serializer->priv->network), NULL);
	}

	g_slist_free (properties);

	// Generate templates
	CpgGroup *template_group = cpg_network_get_template_group (serializer->priv->network);
	GSList const *list = cpg_group_get_children (template_group);
	xmlNodePtr templates;

	if (list)
	{
		templates = xmlNewDocNode (doc, NULL, (xmlChar *)"templates", NULL);
		xmlAddChild (nnetwork, templates);

		group_children_to_xml (serializer, templates, template_group);
	}

	// Generate state and link nodes
	if (serializer->priv->root)
	{
		group_to_xml (serializer, nnetwork, CPG_GROUP (serializer->priv->root));
	}
	else
	{
		group_to_xml (serializer, nnetwork, CPG_GROUP (serializer->priv->network));
	}

	write_extra_nodes (serializer);

	xmlIndentTreeOutput = 1;

	xmlChar *mem;
	int size;

	xmlDocDumpFormatMemoryEnc (doc,
	                           &mem,
	                           &size,
	                           xmlGetCharEncodingName (XML_CHAR_ENCODING_UTF8),
	                           1);

	gboolean ret;

	ret = g_output_stream_write_all (stream,
	                                 mem,
	                                 size,
	                                 NULL,
	                                 NULL,
	                                 error);

	xmlFree (mem);
	xmlFreeDoc (doc);

	return ret;
}

/**
 * cpg_network_serializer_serialize_file:
 * @serializer: A #CpgNetworkSerializer
 * @file: A #GFile
 * @error: A #GError
 * 
 * Convenience function to serialize to a file.
 *
 * Returns: %TRUE if the serialization was successful, %FALSE otherwise
 *
 **/
gboolean
cpg_network_serializer_serialize_file (CpgNetworkSerializer  *serializer,
                                       GFile                 *file,
                                       GError               **error)
{
	g_return_val_if_fail (CPG_IS_NETWORK_SERIALIZER (serializer), FALSE);
	g_return_val_if_fail (G_IS_FILE (file), FALSE);

	GFileOutputStream *stream = g_file_create (file,
	                                           G_FILE_CREATE_NONE,
	                                           NULL,
	                                           error);

	if (!stream)
	{
		return FALSE;
	}

	gboolean ret;

	ret = cpg_network_serializer_serialize (serializer,
	                                        G_OUTPUT_STREAM (stream),
	                                        error);

	g_object_unref (stream);

	return ret;
}

/**
 * cpg_network_serializer_serialize_path:
 * @serializer: A #CpgNetworkSerializer
 * @path: The file path
 * @error: A #GError
 * 
 * Convenience function to serialize to a file path.
 *
 * Returns: %TRUE if the serialization was successful, %FALSE otherwise
 *
 **/
gboolean
cpg_network_serializer_serialize_path (CpgNetworkSerializer  *serializer,
                                       gchar const           *path,
                                       GError               **error)
{
	g_return_val_if_fail (CPG_IS_NETWORK_SERIALIZER (serializer), FALSE);
	g_return_val_if_fail (path != NULL, FALSE);

	GFile *file = g_file_new_for_path (path);

	gboolean ret;

	ret = cpg_network_serializer_serialize_file (serializer,
	                                             file,
	                                             error);

	g_object_unref (file);

	return ret;
}

/**
 * cpg_network_serializer_serialize_memory:
 * @serializer: A #CpgNetworkSerializer
 * @error: A #GError
 *
 * Convenience function to serialize a network to a string.
 *
 * Returns: The serialized network or %NULL if an error occurred.
 *
 **/
gchar *
cpg_network_serializer_serialize_memory (CpgNetworkSerializer  *serializer,
                                         GError               **error)
{
	g_return_val_if_fail (CPG_IS_NETWORK_SERIALIZER (serializer), NULL);

	GOutputStream *stream = g_memory_output_stream_new (NULL,
	                                                    0,
	                                                    g_realloc,
	                                                    NULL);

	gboolean ret;

	ret = cpg_network_serializer_serialize (serializer,
	                                        stream,
	                                        error);

	gpointer data = g_memory_output_stream_get_data (G_MEMORY_OUTPUT_STREAM (stream));
	g_object_unref (stream);

	if (!ret)
	{
		g_free (data);
		data = NULL;
	}

	return data;
}
